"""
This module contains different Models
"""
from abc import ABC, abstractmethod
from typing import Any, AnyStr, Dict

import mph
import pandas as pd
import numpy as np

from ._utils import make_unique


class Model(ABC):
    """
    The Model abstract base class.
    """
    @abstractmethod
    def results(self, *args, **kwargs) -> Any:
        """
        Function to be solved
        """


class TestModel(Model):
    """
    Simple subclass of Model fot tests
    """
    def results(self, x):
        """Returns squared number

        :param x: any number
        :return: squared number
        """
        if x == 1:
            return x
        elif x == 0:
            return x
        raise ValueError('Test value error')


# TODO SMUTHI Model
# TODO logging here?

# noinspection PyMissingOrEmptyDocstring
class ComsolModel(mph.Model, Model):
    def __init__(self):
        super().__init__(None)

    def add_circle(
            self,
            name: str,
            x_i: float,
            y_j: float,
            geometry: mph.Node,
            r: float,
            alpha: float = 1.1
    ):
        node = geometry.create("Circle", name=name)
        node.property("r", str(r))
        node.property("pos", [str(x_i), str(y_j)])

        node_sel = (self/'selections').create('Box', name=name)
        node_sel.property('entitydim', 2)

        modified_r = r * alpha
        node_sel.property('xmin', f'+{x_i}-{modified_r}')
        node_sel.property('xmax', f'+{x_i}+{modified_r}')
        node_sel.property('ymin', f'+{y_j}-{modified_r}')
        node_sel.property('ymax', f'+{y_j}+{modified_r}')
        node_sel.property('condition', 'inside')

        return node, node_sel

    def add_square(
            self, name: str, x_i: float, y_j: float, geometry: mph.Node, width: float, alpha: float = 1.1):
        node = geometry.create("Square", name=name)
        node.property("size", str(width))
        node.property("base", "center")
        node.property("pos", [str(x_i), str(y_j)])

        node_sel = (self/'selections').create('Box', name=name)
        node_sel.property('entitydim', 2)

        modified_width = width / 2 * alpha
        node_sel.property('xmin', f'+{x_i}-{modified_width}')
        node_sel.property('xmax', f'+{x_i}+{modified_width}')
        node_sel.property('ymin', f'+{y_j}-{modified_width}')
        node_sel.property('ymax', f'+{y_j}+{modified_width}')
        node_sel.property('condition', 'inside')

        return node, node_sel

    def add_cylinder(self, name: str,
                     x_i: float, y_j: float, z_k: float,
                     geometry: mph.Node,
                     h: float,
                     r: float, alpha: float = 1.1):

        node = geometry.create("Cylinder", name=name)
        node.property("r", str(r))
        node.property("pos", [str(x_i), str(y_j), str(z_k)])
        node.property("h", str(h))

        node_sel = (self/'selections').create('Box', name=name)
        node_sel.property('entitydim', 3)

        modified_r = r * alpha
        modified_h = h * alpha
        node_sel.property('xmin', f'+{x_i}-{modified_r}')
        node_sel.property('xmax', f'+{x_i}+{modified_r}')
        node_sel.property('ymin', f'+{y_j}-{modified_r}')
        node_sel.property('ymax', f'+{y_j}+{modified_r}')
        node_sel.property('zmin', f'+{z_k}-{modified_h - h}')
        node_sel.property('zmax', f'+{z_k}+{modified_h}')
        node_sel.property('condition', 'inside')

        return node, node_sel

    @staticmethod
    def _clean(node: mph.Node, tag: str):
        for c in node.children():
            if tag in c.path[-1]:
                c.remove()

    def clean_geometry(self, geometry: mph.Node, tag):
        self._clean(geometry, tag)
        self._clean(self/'selections', tag)
        self.build(geometry)

    @staticmethod
    def global_evaluation(dataset: mph.Node, evaluation: mph.Node) -> pd.DataFrame:
        #  https://github.com/MPh-py/MPh/blob/2b967b77352f9ce7effcd50ad4774bf5eaf731ea/mph/model.py#L425
        evaluation.property('data', dataset)
        java = evaluation.java
        real, imag = java.computeResult()
        results = np.array(real) + 1j * np.array(imag)
        return pd.DataFrame(data=results, columns=make_unique(evaluation.property('descr')))

    def clear(self):
        # super().clear()
        super().reset()

    def export_image(self, source: mph.Node, filepath: AnyStr, props: Dict = None):
        exports = self / 'exports'
        image = exports.create('Image')
        default_props = {
            'size': 'manualweb',
            'unit': 'px',
            'height': '720',
            'width': '720'
        }
        for prop in default_props:
            image.property(prop, default_props[prop])
        if props is not None:
            for prop in props:
                image.property(prop, props[prop])

        image.property('sourceobject', source)
        image.property('filename', filepath)

        self.export()
        image.remove()

    def plot2d(self, expr: AnyStr, filepath: AnyStr, props: Dict = None):
        plots = self / 'plots'
        plots.java.setOnlyPlotWhenRequested(True)
        plot = plots.create('PlotGroup2D')

        surface = plot.create('Surface', name='plot2d')
        surface.property('resolution', 'normal')
        surface.property('expr', expr)

        self.export_image(plot, filepath, props)
        plot.remove()

    def getLastComputationTime(self):
        # TODO ADD ANY STUDY SUPPORT

        studies = (self / 'studies').children()
        return -1 if not len(studies) else int(studies[-1].java.getLastComputationTime())

    @abstractmethod
    def results(self, *args: Any, **kwargs: Any) -> Any:
        pass

    def configure(self) -> Any:
        pass

    def pre_build(self, *args: Any, **kwargs: Any):
        pass

    def pre_solve(self, *args: Any, **kwargs: Any):
        pass

    def pre_clear(self, *args: Any, **kwargs: Any):
        pass



"""
This module is used by solvers and models mostly
"""
from typing import Tuple, List, Sequence, Dict
import numpy as np
from .task import Task


def make_unique(labels: Sequence) -> list:
    """Gets labels list and makes all labels unique by adding '(number of inclusion)' postfix

    :param labels: List of labels
    :return: List of renamed labels
    """
    new_labels = []
    _, real_index, counts = np.unique(labels, return_counts=True, return_inverse=True)
    for index in range(len(labels)):
        for count in range(counts[real_index[index]]):
            new_label = labels[index] + (f'({count})' if count != 0 else '')
            if new_label not in new_labels:
                new_labels.append(new_label)
                break

    return new_labels


def tasks_sort(tasks: Sequence[Task], cache: Dict) -> Tuple[List[int], List[int], List[int]]:
    """Sorts the tasks list into:
        * those that need to be solved
        * those that are already in the cache
        * those that are repeated in the tasks list
    and returns three lists of task indexes

    :param tasks: List of tasks
    :param cache: Cache dict
    :return: tasks indexes sorted into three lists
    :raises ValueError: raises if any task does not contain
    tag property even if caching is used on solver
    """
    to_solve, cached, same = [], [], []
    to_solve_tags = []
    for i, task in enumerate(tasks):
        if task.tag is None:
            raise ValueError('If caching is True all the tasks must have a tag property')
        if task.tag in cache:
            cached.append(i)
        elif task.tag in to_solve_tags:
            same.append(i)
        else:
            to_solve.append(i)
            to_solve_tags.append(task.tag)
    return to_solve, cached, same



"""
This module contains Task class
"""
from typing import Hashable


class Task:
    """Task class"""
    def __init__(self, *args, tag=None, **kwargs):
        if not isinstance(tag, Hashable):
            raise TypeError('tag must be Hashable')
        self.tag = tag
        self.args = args
        self.kwargs = kwargs



"""
A simple package for parallel computing with Python
"""
from .models import (
    Model,
    TestModel,
    ComsolModel
)
from .workers import (
    Worker,
    ComsolMultiprocessingWorker,
    ComsolWorker,
    SimpleWorker,
    SimpleMultiprocessingWorker,
    MPIWorker,
    ComsolMPIWorker,
    SimpleMPIWorker
)
from .solvers import (
    MultiprocessingSolver,
    SimpleSolver,
    MPISolver,
    Solver,
    get_solver
)
from .task import Task



"""
This module contains different Solvers
"""
import multiprocessing
import sys
from abc import ABC, abstractmethod

from multiprocessing import Queue, JoinableQueue
from typing import Any, Sequence, Union, List, Iterator, Callable, Iterable
import time
import logging

from .task import Task
from .workers import Worker, MultiprocessingWorker, MPIWorker
from ._utils import tasks_sort

# TODO(add info to return after .solve())
# TODO(add additional threads to control workers' errors)

logger = logging.getLogger(__package__ + '.solver')


def task_iterator(iterable: Sequence[Task]) -> Iterator[Task]:
    """Returns Iterator from on iterable"""
    return iterable.__iter__()


class Solver(ABC):
    """Base Solver class"""
    def __init__(self):
        self.caching = False
        self.cache = {}
        self.total_workers = 0

    def solve(self,
              tasks: Sequence[Task],
              iterator: Callable[[Sequence[Task]], Union[Iterator[Task], Iterable[Task]]] = task_iterator
              ) -> List[Any]:
        """Solves tasks

        :param tasks: tasks to solve
        :param iterator: function which wraps iterable tasks list and return iterator
        :return: list of results
        """
        if self.caching:
            to_solve, cached, same = tasks_sort(tasks, self.cache)
        else:
            to_solve, cached, same = range(len(tasks)), [], []

        message = f'Starting to solve {len(tasks)} tasks with {self.total_workers} workers'
        if self.caching:
            message += f':\n\t{len(cached)} solutions will be reused\n' \
                       f'\t{len(same)} tasks are the same'
        else:
            message += ' (caching is off)'
        logger.info(message)
        start_time = time.time()

        res = self._solve(tasks, to_solve, cached, same, iterator)

        end_time = time.time()
        logger.info('All the tasks have been solved in %.2fs', end_time - start_time)
        return res

    @abstractmethod
    def _solve(
            self,
            tasks: Sequence[Task],
            to_solve: List[int],
            cached: List[int],
            same: List[int],
            iterator: Callable[[Sequence[Task]], Iterator[Task]]
    ) -> List[Any]:
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class MultiprocessingSolver(Solver):
    """Multiprocessing Solver implementation"""
    def __init__(self,
                 worker: MultiprocessingWorker,
                 workers_num: int = 1,
                 caching: bool = False
                 ):
        super().__init__()
        self.caching = caching
        self.worker = worker
        if workers_num <= 0:
            raise RuntimeError('At least 1 worker is needed')
        self.total_workers = workers_num

        self.workers = []

        self._jobs = JoinableQueue()
        self._results = Queue()

        for _ in range(self.total_workers):
            process = multiprocessing.Process(
                target=self.worker.start_loop,
                args=(self._jobs, self._results),
                daemon=True
            )
            process.start()
            self.workers.append(process)

    def _solve(
            self,
            tasks: Sequence[Task],
            to_solve: List[int],
            cached: List[int],
            same: List[int],
            iterator: Callable[[Sequence[Task]], Iterator[Task]]
    ) -> List[Any]:
        results = [None for _ in tasks]
        for i, task in enumerate(tasks):
            if i in to_solve:
                self._jobs.put((i, task.args, task.kwargs))
            elif i in cached:
                results[i] = self.cache[task.tag]
            else:
                results[i] = None
        tasks_to_solve = [task for i, task in enumerate(tasks) if i in to_solve]
        for _ in iterator(tasks_to_solve):
            (i, res) = self._results.get()
            if i == -1:
                raise RuntimeError(res)
            if self.caching:
                self.cache[tasks[i].tag] = res
            results[i] = res

        for i in same:
            results[i] = self.cache[tasks[i].tag]

        return results

    def _stop(self):
        for worker in self.workers:
            if worker.is_alive():
                worker.terminate()
                worker.join()
                worker.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop()
        return False


class SimpleSolver(Solver):
    """Simplest Solver implementation"""
    def __init__(self, worker: Worker, caching: bool = False):
        if not isinstance(worker, Worker):
            raise TypeError('Worker has to be the object of type Worker')
        super().__init__()
        self.caching = caching
        self.worker = worker
        self.worker.start()
        self.total_workers = 1

    def _solve(
            self,
            tasks: Sequence[Task],
            to_solve: List[int],
            cached: List[int],
            same: List[int],
            iterator: Callable[[Sequence[Task]], Iterator[Task]]
    ) -> List[Any]:
        results = [None for _ in tasks]
        for i, task in enumerate(iterator(tasks)):
            if i in to_solve:
                try:
                    results[i] = self.worker.do_the_job(task.args, task.kwargs)
                except Exception as err:
                    raise RuntimeError from err
                if self.caching:
                    self.cache[task.tag] = results[i]
            elif i in cached or i in same:
                results[i] = self.cache[task.tag] if task.tag is not None else None
            else:
                results[i] = None
        return results


class MPISolver(Solver):
    """MPI Solver implementation"""
    def __init__(self, worker: MPIWorker, caching: bool = False, buffer_size: int = 32768):
        from mpi4py import MPI
        if not isinstance(worker, MPIWorker):
            raise TypeError('Worker has to be the object of type MPIWorker')
        super().__init__()
        self.worker = worker
        self.caching = caching
        self.buffer_size = buffer_size
        self.comm = MPI.COMM_WORLD
        self.rank = self.comm.Get_rank()
        self.total_workers = self.comm.Get_size() - 1

        if self.rank != 0:
            logger.debug('Starting loop in worker with rank %i', self.comm.Get_rank())
            self.worker.start_loop(self.comm)
            # self.start_listening()
            MPI.Finalize()
            sys.exit()

        if self.total_workers < 1:
            raise RuntimeError('Not enough processes! '
                               f'Need at least 2, detected {self.total_workers}.')

    def _solve(
            self,
            tasks: Sequence[Task],
            to_solve: List[int],
            cached: List[int],
            same: List[int],
            iterator: Callable[[Sequence[Task]], Iterator[Task]]
    ) -> List[Any]:
        results = [None for _ in tasks]
        requests = []
        tasks_to_solve = []
        for i, task in enumerate(tasks):
            if i in to_solve:
                dest = len(requests) % self.total_workers + 1
                req = self.comm.isend((i, task.args, task.kwargs), dest=dest, tag=i)
                req.wait()
                requests.append(self.comm.irecv(self.buffer_size, source=dest))
                tasks_to_solve.append(task)
            elif i in cached:
                results[i] = self.cache[task.tag]
        for request, task in zip(requests, iterator(tasks_to_solve)):
            i, res = request.wait()
            if self.caching:
                self.cache[tasks[i].tag] = res
            results[i] = res

        for i in same:
            results[i] = self.cache[tasks[i].tag]

        return results

    def _stop(self):
        for i in range(1, self.comm.Get_size()):
            req = self.comm.isend((None, None, None), dest=i)
            req.wait()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop()


def get_solver(
        worker: Union[Worker, MPIWorker, MultiprocessingWorker],
        caching: bool = False,
        workers_num: int = 1,
        buffer_size: int = 2**15,
) -> Union[SimpleSolver, MPISolver, MultiprocessingSolver]:
    """Automatically choose solver based on worker type.

    :param worker: worker instance
    :param caching: is caching enabled
    :param workers_num: number of workers to spawn if worker is instance of MultiprocessingWorker
    :param buffer_size: buffer size for messages with results used by MPI
    :return: solver
    """
    if isinstance(worker, MultiprocessingWorker):
        return MultiprocessingSolver(worker=worker, workers_num=workers_num, caching=caching)
    if isinstance(worker, MPIWorker):
        return MPISolver(worker, caching, buffer_size)
    if isinstance(worker, Worker):
        return SimpleSolver(worker, caching)
    raise NotImplementedError(f'can\'t get Solver for {type(worker)} worker type')



"""
This module contains different Workers
"""
import multiprocessing
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple
from multiprocessing import JoinableQueue, Queue
from pathlib import Path
import socket
import logging
import mph

from .models import ComsolModel, Model

logger = logging.getLogger(__package__ + '.worker')


class Worker(ABC):
    """
    The Worker abstract base class.
    """
    @abstractmethod
    def start(self):
        """Function that stars the worker. Always called on each worker by solver when initializing.

        :return:
        """

    @abstractmethod
    def do_the_job(self, args: Tuple[Any], kwargs: Dict[str, Any]) -> Any:
        """calls Model.results() with specified params

        :param args: args for Model.results()
        :param kwargs: kwargs for Model.results()
        :return: pass Model.results()
        """


class MultiprocessingWorker(Worker, ABC):
    """Class of basic multiprocessing worker."""
    def start_loop(self, jobs: JoinableQueue, results: Queue):
        """Loop starting will be called by multiprocessing solver.

        :param JoinableQueue jobs: Queue for jobs
        :param Queue results: Queue for results
        """
        logger.info('%s %s Starting',
                    multiprocessing.current_process().name,
                    socket.gethostname())
        self.start()
        self._loop(jobs, results)

    def _loop(self, jobs: JoinableQueue, results: Queue):
        logger.info('%s %s Entering the loop',
                    multiprocessing.current_process().name,
                    socket.gethostname())
        while True:
            (i, args, kwargs) = jobs.get()
            logger.debug('%s %s Starting doing the job %i',
                         multiprocessing.current_process().name,
                         socket.gethostname(),
                         i)
            try:
                res = self.do_the_job(args, kwargs)
                results.put((i, res))
                jobs.task_done()
            except Exception as err:
                results.put((-1, err))
                raise err


class MPIWorker(Worker, ABC):
    """Class of basic MPI worker"""
    def start_loop(self, comm):
        """Loop start will be called by MPI solver.

        :param comm: Intracomm, COMM_WORLD commonly
        """
        import mpi4py
        if not isinstance(comm, mpi4py.MPI.Intracomm):
            raise TypeError('comm has to be instance of mpi4py.MPI.Intracomm')
        logger.info('Rang %i %s Starting',
                    comm.Get_rank(),
                    socket.gethostname())
        self.start()
        self._loop(comm)

    def _loop(self, comm):
        logger.info('Rang %i %s Entering the loop',
                    comm.Get_rank(),
                    socket.gethostname())
        while True:
            req = comm.irecv(source=0)
            (i, args, kwargs) = req.wait()
            if i is None and args is None and kwargs is None:
                return
            logger.debug('Rang %i %s Starting doing the job %i',
                         comm.Get_rank(),
                         socket.gethostname(),
                         i)
            res = self.do_the_job(args, kwargs)
            req = comm.isend((i, res), dest=0)
            req.wait()


class SimpleWorker(Worker):
    """Simplest implementation of worker"""
    def __init__(self, model: Model):
        self.model = model

    def start(self):
        """Start of worker"""

    def do_the_job(self, args: Tuple[Any], kwargs: Dict[str, Any]) -> Any:
        """Get model results"""
        return self.model.results(*args, **kwargs)


class SimpleMultiprocessingWorker(SimpleWorker, MultiprocessingWorker):
    """Simplest multiprocessing worker"""


class SimpleMPIWorker(SimpleWorker, MPIWorker):
    """Simplest MPI worker"""


# noinspection PyMissingOrEmptyDocstring
class ComsolWorker(Worker):
    def __init__(self, model: ComsolModel, filepath: str or Path,
                 mph_options: dict = None,
                 client_args=None,
                 client_kwargs=None):
        if not isinstance(model, ComsolModel):
            raise TypeError('Model has to be an object of ComsolModel class')

        self.client = None
        self.model = model

        self._mph_options = {} if mph_options is None else mph_options
        self._client_args = [] if client_args is None else client_args
        self._client_kwargs = {} if client_kwargs is None else client_kwargs
        self._filepath = filepath

    def start(self):
        for option in self._mph_options:
            mph.option(option, self._mph_options[option])
        logger.debug(f'Opening COMSOL model {self._filepath}')
        self.client = mph.start(*self._client_args, **self._client_kwargs)  # type: mph.client
        self.model.java = self.client.load(self._filepath).java
        self.model.configure()

    def do_the_job(self, args, kwargs) -> Any:
        self.model.pre_build(*args, **kwargs)
        self.model.build()
        self.model.pre_solve(*args, **kwargs)
        self.model.mesh()
        self.model.solve()
        results = self.model.results(*args, **kwargs)
        self.model.pre_clear(*args, **kwargs)
        self.model.clear()
        return results


# noinspection PyMissingOrEmptyDocstring
class ComsolMultiprocessingWorker(ComsolWorker, MultiprocessingWorker):
    def start(self):
        super().start()


# noinspection PyMissingOrEmptyDocstring
class ComsolMPIWorker(ComsolWorker, MPIWorker):
    def start(self):
        super().start()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pckit']

package_data = \
{'': ['*']}

install_requires = \
['MPh>=1.1.5,<2.0.0',
 'mpi4py>=3.1.3,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'pckit',
    'version': '0.2.0',
    'description': 'A simple package for parallel computing with Python',
    'long_description': "# pckit\n\nThis is a simple package for parallel computing with Python.\n\n## Usage\n### Multiprocessing\nSimple multiprocess solver usage.\nHere the square of 2 and 3 are evaluated by 2 workers.\n\n```python\nimport pckit.task\nimport pckit\n\n\nclass MyModel(pckit.Model):\n    def results(self, x):\n        # solve here your problem f(x) = x^2\n        return x ** 2\n\n\nif __name__ == '__main__':\n    model = MyModel()\n    worker = pckit.SimpleMultiprocessingWorker(model)\n    with pckit.get_solver(worker, workers_num=2) as solver:\n        # create tasks to solve. You can put args or kwargs here\n        tasks = [pckit.task.Task(2), pckit.task.Task(x=3)]\n        results = solver.solve(tasks)\n        print(results)\n        # >>> [4, 9]\n```\n\n### MPI\nYou can simply run scripts on the cluster with [mpi4py](https://github.com/mpi4py/mpi4py) implementation on MPI. \nSimply by changing `SimpleMultiprocessingWorker` to `SimpleMPIWorker` in previous example and starting script with MPI `mpiexec -np 3 python -m mpi4py your_script.py`.\n\n```python\nworker = pckit.SimpleMPIWorker(model)\n```\n\n### Examples\n[More examples](./examples)\n\n## Features\n### Cache\nDict based cache is available by `caching` argument in solver.\n`tag` property in `Task` is required and has to be Hashable.\n\n```python\nwith pckit.get_solver(worker, caching=True) as solver:\n    tasks = [pckit.Task(2, tag='2'), pckit.Task(2, tag='2')]\n```\nSecond result will be reused from cache.",
    'author': 'djiboshin',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/djiboshin/pckit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
